/*
 * mod_suid2.c for Apache-2.0
 *
 * Copyright 2004 by Hideo NAKAMITSU. All rights reserved
 *
 * Hideo NAKAMITSU <nomo@bluecoara.net>
 * http://bluecoara.net/ (Japanese)
 *
 */

/*
 * --------------
 * WHAT IS THIS ?
 * --------------
 *
 * With this module, all httpd process run under user's access right, not nobody or apache.
 * mod_suid2 is similar to mod_suexec, but mod_suexec can only work with CGIs or SSIs.
 * If you want to run apache modules, i.e. WebDAV, PHP, and so on under user's right,
 * this module is useful.
 *
 * This is for ISP or ASP which has a large number of users or virtualhosts.
 *
 * ------
 * HOW TO
 * ------
 *
 * 1. Build Apache-2.0.x with -DBIG_SECURITY_HOLE.
 * i.e.
 * % CFLAGS="-DBIG_SECURITY_HOLE" ./configure && make
 *
 * 2. Install this module.
 *
 * 3. Set User directive to root in your httpd.conf.
 *    Because child process run as user's right by setuid().
 *
 * 4. Set "SuidUserGroup Username Groupname" or "SuidUidGid uidNumber gidNumber"
 *    in you httpd.conf. Configuration  format is like SuexecUserGroup.
 *
 * 5. Restart your Apache.
 *
 * 6. Enjoy.
 *
 * -------
 * INSTALL
 * -------
 *
 * Very easy.
 *
 * # /usr/local/apache2/bin/apxs -a -i -c mod_suid2.c
 *
 * -------------
 * CONFIGURATION
 * -------------
 *
 * Format is
 *
 *   SuidUserGroup Username Groupname
 *
 * or
 *
 *   SuidUserGroup #uidNumber #gidNumber
 *
 * or
 *
 *   SuidUidGid uidNumber gidNumber
 *
 * That Username(uidNumber) or Groupname(gidNumber) must be upper than UID_MIN(default = 100) and GID_MIN(default = 100)
 *
 * -------
 * EXAMPLE
 * -------
 *
 * 1. Server section
 *
 * LoadModule suid2_module  modules/mod_suid2.so
 * User                     root
 * Group                    root
 * SuidUserGroup            apache apache # Default User and Group (if process is not under Virtualhost)
 *
 * 2. VirtualHost section
 *
 * NameVirtualHost 192.168.0.1
 * <VirtualHost example.com>
 *   ServerAdmin    webmaster@example.com
 *   DocumentRoot   /home/example.com/public_html
 *   ServerName     example.com
 *   ServerAlias    www.example.com
 *   SuidUserGroup  example1 example1
 * </VirtualHost>
 *
 * <VirtualHost example.net>
 *   ServerAdmin    webmaster@example.net
 *   DocumentRoot   /home/example.net/public_html
 *   ServerName     example.net
 *   ServerAlias    www.example.net
 *   SuidUserGroup  example2 example2
 * </VirtualHost>
 *
 * Then, httpd process of "example.com" will run as "example1",
 * "example.net" will be "example2".
 *
 * Another configuration example.
 * This is useful when a user doesn't have Username or Groupname.
 *   SuidUserGroup #1001 #1001        # run httpd under uidNumber=1001, gidNumber=1001
 *   SuidUidGid     1001  1001        # run httpd under uidNumber=1001, gidNumber=1001
 * Notice, the maximum uid or gid number depends on your OS.
 *
 * -----------
 * PERFORMANCE
 * -----------
 *
 * To setuid() and setgid() httpd processes, child process must be running as root all the time.
 * It means once child process is running under user's right, that can not change to another user,
 * because only ROOT can change process rights.
 *
 * The problem is one child process will killed after only one request.
 * To do this, mod_suid2 sets MaxRequestsPerChild to 1 in the program.
 *
 * Setting MaxRequestsPerChild to 1, performance will be down as you know.
 * If you need highperformance, Modify your httpd.conf like this.
 * But this is slowly than normal apache2.
 * 
 * StartServers          50
 * MinSpareServers       50
 * MaxSpareServers      100
 * MaxClients           150
 * MaxRequestsPerChild    0
 *
 * --------
 * SECURITY
 * --------
 *
 * !! AT YOUR OWN RISK !!
 *
 * To enable this module, you have to compile apache2 with -DBIG_SECURITY_HOLE.
 * If your apache is old, Some Exploits may be very dangerous.
 *
 * To secure Apache, there are some ways.
 *
 * 1. Install IDS and IPS, like Snort or another to detect or protect attacks.
 *
 * 2. Run your apache under chroot or jail, to protect root file systems.
 *
 * 3. Install libsafe or compile your apache with StackGuard, to prevent buffer over flow.
 *
 * 4. Upgrade all software to highest version.
 *
 * ----
 * TEST
 * ----
 *
 * You can test setting by some CGIs like this
 *
 * #!/bin/sh
 * printf "Content-type: text/html\n\n";
 * /usr/bin/whoami
 *
 */
#include "apr_strings.h"
#include "apr_md5.h"
#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "mpm_common.h"

#define MODULE                "mod_suid2"
#define SUID_DEFAULT_USER     "nobody"
#define SUID_DEFAULT_GROUP    "nobody"
#define UID_MIN               100
#define GID_MIN               100
#define NDEBUG	              1

typedef struct
{
  int suid2_uid;
  int suid2_gid;
  int suid2_def_uid;
  int suid2_def_gid;
} suid2_config_t;

module AP_MODULE_DECLARE_DATA suid2_module;

static void *
create_config (apr_pool_t * p, char *d)
{
  suid2_config_t *conf = apr_palloc (p, sizeof (*conf));
  conf->suid2_uid = conf->suid2_def_uid = ap_uname2id (SUID_DEFAULT_USER);	/* Default User */
  conf->suid2_gid = conf->suid2_def_gid = ap_gname2id (SUID_DEFAULT_GROUP);	/* Default Group */
  return conf;
}

static const char *
set_suid2_usergroup (cmd_parms * cmd, void *mconfig,
		     const char *uid, const char *gid)
{
  suid2_config_t *conf = ap_get_module_config (cmd->server->module_config,
					       &suid2_module);
  const char *err =
    ap_check_cmd_context (cmd, NOT_IN_DIR_LOC_FILE | NOT_IN_LIMIT);
  if (err != NULL)
    {
      return err;
    }
  conf->suid2_uid = ap_uname2id (uid);
  conf->suid2_gid = ap_gname2id (gid);
  return NULL;
}

static const char *
set_suid2_uidgid (cmd_parms * cmd, void *mconfig,
		  const char *uid, const char *gid)
{
  suid2_config_t *conf = ap_get_module_config (cmd->server->module_config,
					       &suid2_module);
  const char *err =
    ap_check_cmd_context (cmd, NOT_IN_DIR_LOC_FILE | NOT_IN_LIMIT);
  if (err != NULL)
    {
      return err;
    }
  conf->suid2_uid = atoi (uid);
  conf->suid2_gid = atoi (gid);
  return NULL;
}

static const command_rec suid2_cmds[] = {
  /* configuration of httpd.conf */
  AP_INIT_TAKE2 ("SuidUserGroup", set_suid2_usergroup, NULL, RSRC_CONF,
		 "User and group for spawned processes"),
  AP_INIT_TAKE2 ("SuidUidGid", set_suid2_uidgid, NULL, RSRC_CONF,
		 "uidNumber and gidNumber for spawned processes"),
  {NULL}
};

static apr_status_t
suid2_exit (void *data)
{
#ifndef NDEBUG
  /* Debug only (when child process killed) */
  ap_log_error (APLOG_MARK, APLOG_DEBUG, 0, NULL,
		"%s child pid=%ld dead", MODULE, (long) getpid ());
#endif
}

static int
suid2_init (apr_pool_t * p, server_rec * s)
{
  if (getuid () == 0)
    {
      /*
       * MaxRequestsPerChild MUST be 1
       * to kill child which already setgid() and setuid()
       */
      ap_max_requests_per_child = 1;

      /*
       * setegid() and seteuid() for the secirity.
       */
      setegid (ap_gname2id (SUID_DEFAULT_GROUP));
      seteuid (ap_uname2id (SUID_DEFAULT_USER));
#ifndef NDEBUG
      ap_log_error (APLOG_MARK, APLOG_DEBUG, 0, NULL,
		    "%s child pid=%ld started", MODULE, (long) getpid ());
#endif
    }
  else
    {
      ap_log_error (APLOG_MARK, APLOG_ERR, 0, NULL,
		    "%s User must be root. %s doest not work correctly",
		    MODULE, MODULE);
    }
  /* For child process exit */
  apr_pool_cleanup_register (p, s, suid2_exit, suid2_exit);
  return OK;
}

static int
suid2_access (request_rec * r)
{
  suid2_config_t *conf = ap_get_module_config (r->server->module_config,
					       &suid2_module);
  /* setgid & setuid */
  if (conf->suid2_gid < GID_MIN && conf->suid2_gid != conf->suid2_def_gid)
    {
      ap_log_error (APLOG_MARK, APLOG_ERR, 0, NULL,
		    "%s %s %s GID(%d) < GID_MIN(%d)",
		    MODULE, ap_get_server_name (r), r->the_request,
		    conf->suid2_gid, GID_MIN);
      return HTTP_FORBIDDEN;
    }
  if (conf->suid2_uid < UID_MIN && conf->suid2_uid != conf->suid2_def_uid)
    {
      ap_log_error (APLOG_MARK, APLOG_ERR, 0, NULL,
		    "%s %s %s UID(%d) < UID_MIN(%d)",
		    MODULE, ap_get_server_name (r), r->the_request,
		    conf->suid2_gid, UID_MIN);
      return HTTP_FORBIDDEN;
    }

  /*
   * seteuid(0) before setuid(0)
   */
  seteuid (0);
  setegid (0);
  if (setgid (conf->suid2_gid) != 0)
    {
      ap_log_error (APLOG_MARK, APLOG_ERR, 0, NULL,
		    "%s %s %s setgid(%d) failed. getgid=%d",
		    MODULE, ap_get_server_name (r), r->the_request,
		    conf->suid2_gid, getgid ());
      return HTTP_FORBIDDEN;
    }
  if (setuid (conf->suid2_uid) != 0)
    {
      ap_log_error (APLOG_MARK, APLOG_ERR, 0, NULL,
		    "%s %s %s setuid(%d) failed. getuid=%d",
		    MODULE, ap_get_server_name (r), r->the_request,
		    conf->suid2_uid, getuid ());
      return HTTP_FORBIDDEN;
    }
#ifndef NDEBUG
  ap_log_error (APLOG_MARK, APLOG_DEBUG, 0, NULL,
		"%s %s %s setgid(%d) setuid(%d) successful",
		MODULE, ap_get_server_name (r), r->the_request, getgid (),
		getuid ());
#endif
  return OK;
}

static void
register_hooks (apr_pool_t * p)
{
  ap_hook_child_init (suid2_init, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_access_checker (suid2_access, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA suid2_module = {
  STANDARD20_MODULE_STUFF,
  NULL,				/* dir config creater */
  NULL,				/* dir merger --- default is to override */
  create_config,		/* server config */
  NULL,				/* merge server config */
  suid2_cmds,			/* command apr_table_t */
  register_hooks		/* register hooks */
};
